use strict;
use File::Spec;
use File::Find;

my %removed_funcs = (
    'map' => 'fn:apply',                   # leave off optional fn: namespace
    'map-pairs' => 'fn:for-each-pair',     # leave off optional fn: namespace
    'map:for-each-entry' => 'map:for-each',
    'map:new' => 'map:merge',
    'util:catch' => 'XQuery 3.1 try-catch expression',
    'util:eval-async' => '???',
    'util:serialize' => 'fn:serialize',
    'validation:validate' => 'more specific validation function',
    'validation:validate-report' => 'more specific validation function',
    'xmldb:add-user-to-group' => 'sm:add-group-member',
    'xmldb:change-user' => 'more specific sm function',
    'xmldb:chmod-collection' => 'sm:chmod',
    'xmldb:chmod-resource' => 'sm:chmod',
    'xmldb:copy-collection' => 'xmldb:copy',
    'xmldb:copy-resource' => 'xmldb:copy',
    'xmldb:create-group' => 'sm:create-group',
    'xmldb:create-user' => 'sm:create-account',
    'xmldb:delete-user' => 'sm:remove-account',
    'xmldb:document' => 'fn:doc',
    'xmldb:exists-user' => 'sm:user-exists',
    'xmldb:get-current-user' => 'sm:id',
    'xmldb:get-current-user-attribute' => '???',
    'xmldb:get-current-user-attribute-names' => '???',
    'xmldb:get-group' => 'sm:get-group*',
    'xmldb:get-owner' => '???',
    'xmldb:get-user-groups' => 'sm:get-user-groups',
    'xmldb:get-user-home' => '???',
    'xmldb:get-user-primary-group' => 'sm:get-user-primary-group',
    'xmldb:get-users' => 'sm:list-users',
    'xmldb:group-exists' => 'sm:group-exists',
    'xmldb:is-admin-user' => 'sm:is-dba',
    'xmldb:is-authenticated' => 'sm:is-authenticated or sm:is-externally-authenticated',
    'xmldb:get-permissions' => 'sm:get-permissions',
    'xmldb:permissions-to-string' => '???',
    'xmldb:string-to-permissions' => '???',
    'xmldb:remove-user-from-group' => 'sm:remove-group-member',
    'xmldb:set-collection-permissions' => 'sm:chmod sm:chown sm:chgrp',
    'xmldb:set-resource-permissions' => 'sm:chmod sm:chown sm:chgrp',
);

my %removed_modules = (
    'context' => 'http://exist-db.org/xquery/context|obsolete -- no replacement available',
    'datetime' => 'http://exist-db.org/xquery/datetime|XQuery 3.1, FunctX, or other implementations',
    'ftp' => 'http://exist-db.org/xquery/ftpclient|EXPath File Transfer Client',
    'httpclient' => 'http://exist-db.org/xquery/httpclient|EXPath HTTP Client',
    'math-ext' => 'http://exist-db.org/xquery/math|XQuery 3.1 math module',
    'memcached' => 'http://exist-db.org/xquery/memcached|obsolete -- no replacement available',
    'svn' => 'http://exist-db.org/xquery/versioning/svn|https://github.com/shabanovd/eXist-svn',
    'xmpp' => 'http://exist-db.org/xquery/xmpp|obsolete -- no replacement available',
);

my %func_lines;
my %module_lines;

my @files;

sub match_files {
    push @files, $File::Find::name if ($_ =~ m/\.xq(l|m)?$/);
}

my $dir = $ARGV[0] || File::Spec->curdir();
find(\&match_files, $dir);

sub scan_file {
    my $file = shift;
    open my $fh, '<', $file or die "Couldn't open $file: $!";
    my $line_num = 0;
    while (my $line = <$fh>) {
        $line_num++;
        for my $func (sort keys %removed_funcs) {
            if ($line =~ m/(?<![\$\-])\b # word boundary but negative lookbehind preventing $ or -
                ${func}\s*               # function name followed by optional whitespace
                \(\s*                    # left paren followed by optional whitespace
                (?!([\)\+\?\\*]          # negative lookahead disallowing quantifiers
                |xs:integer              # and data types, which indicate declarations
                |xs:string               # rather than functions (mostly for 'map')
                |function
                |xs:anyURI
                |xs:boolean
                |xs:byte
                |xs:date
                |xs:dateTime
                |xs:decimal
                |xs:dayTimeDuration
                |xs:double
                |xs:duration
                |xs:float
                |xs:gDay
                |xs:gMonthDay
                |xs:gYear
                |xs:gYearMonth
                |xs:long
                |xs:Name
                |xs:QName
                |xs:short
                |xs:time
                |xs:yearMonthDuration
                ))/x) {
                    push @{$func_lines{$func}}, "$file:$line_num $line";
            }
        }
        for my $mod (sort keys %removed_modules) {
            my ($ns, $replacement) = split /\|/, $removed_modules{$mod};
            my $namespace = qr{$ns};
            if ($line =~ m/import module namespace\s+.*$namespace/) {
                push @{$module_lines{$mod}}, "$file:$line_num $line";
            }

        }
    }
    close $fh;
}


while (my $file = shift @files) {
    scan_file($file);
}

for my $func (sort keys %func_lines) {
    print "\n>>>  Replace the following instances of the function $func with $removed_funcs{$func}.\n\n";
    for (my $line = shift @{$func_lines{$func}}) { print $line; }
}

for my $mod (sort keys %module_lines) {
    my ($ns, $replacement) = split /\|/, $removed_modules{$mod};
    print "\n>>>  Replace the following instances of the module $mod with $replacement.\n\n";
    for (my $line = shift @{$module_lines{$mod}}) { print $line; }
}

exit;
